#include <stdlib.h>

#include "types.h"
#include "memory.h"
#include "chacha20.h"

static inline u32 rotl32(u32 x, int n) 
{
  return (x << n) | (x >> (32 - n));
}

static inline u32 pack4(const u8 *a)
{
  u32 r = 0;
  
  r |= (u32)a[0] << 0 * 8;
  r |= (u32)a[1] << 1 * 8;
  r |= (u32)a[2] << 2 * 8;
  r |= (u32)a[3] << 3 * 8;

  return r;
}

static inline void unpack4(u32 src, u8 *dst)
{
  dst[0] = (src >> 0 * 8) & 0xFF;
  dst[1] = (src >> 1 * 8) & 0xFF;
  dst[2] = (src >> 2 * 8) & 0xFF;
  dst[3] = (src >> 3 * 8) & 0xFF;
}

static void chacha20_init_block(chacha20_ctx_t *ctx, u8 key[], u8 nonce[])
{
  passman_memcpy(ctx->key, key, sizeof(ctx->key));
  passman_memcpy(ctx->nonce, nonce, sizeof(ctx->nonce));

  const u8 *magic_constant = (u8 *)"expand 32-byte k";
  
  ctx->state[0]  = pack4(magic_constant);
  ctx->state[1]  = pack4(magic_constant + 1 * 4);
  ctx->state[2]  = pack4(magic_constant + 2 * 4);
  ctx->state[3]  = pack4(magic_constant + 3 * 4);
  ctx->state[4]  = pack4(key + 0 * 4);
  ctx->state[5]  = pack4(key + 1 * 4);
  ctx->state[6]  = pack4(key + 2 * 4);
  ctx->state[7]  = pack4(key + 3 * 4);
  ctx->state[8]  = pack4(key + 4 * 4);
  ctx->state[9]  = pack4(key + 5 * 4);
  ctx->state[10] = pack4(key + 6 * 4);
  ctx->state[11] = pack4(key + 7 * 4);

  ctx->state[12] = 0;
  ctx->state[13] = pack4(nonce);
  ctx->state[14] = pack4(nonce + 1 * 4);
  ctx->state[15] = pack4(nonce + 2 * 4);
  
  passman_memcpy(ctx->nonce, nonce, sizeof(ctx->nonce));
}

static void chacha20_block_set_counter(chacha20_ctx_t *ctx, u64 counter)
{
  ctx->state[12] = (u32)counter;
  ctx->state[13] = pack4(ctx->nonce + 0 * 4) + (u32)(counter >> 32);
}

static void chacha20_block_next(chacha20_ctx_t *ctx)
{
  for (u32 i = 0; i < 16; i++)
    ctx->keystream32[i] = ctx->state[i];
  
#define CHACHA20_QUARTERROUND(x, a, b, c, d)	\
  do {						\
  x[a] += x[b]; x[d] = rotl32(x[d] ^ x[a], 16); \
  x[c] += x[d]; x[b] = rotl32(x[b] ^ x[c], 12); \
  x[a] += x[b]; x[d] = rotl32(x[d] ^ x[a], 8);	\
  x[c] += x[d]; x[b] = rotl32(x[b] ^ x[c], 7);	\
  } while (0)
  
  for (u32 i = 0; i < 10; i++) 
    {
      CHACHA20_QUARTERROUND(ctx->keystream32, 0, 4,  8, 12);
      CHACHA20_QUARTERROUND(ctx->keystream32, 1, 5,  9, 13);
      CHACHA20_QUARTERROUND(ctx->keystream32, 2, 6, 10, 14);
      CHACHA20_QUARTERROUND(ctx->keystream32, 3, 7, 11, 15);
      
      CHACHA20_QUARTERROUND(ctx->keystream32, 0, 5, 10, 15);
      CHACHA20_QUARTERROUND(ctx->keystream32, 1, 6, 11, 12);
      CHACHA20_QUARTERROUND(ctx->keystream32, 2, 7,  8, 13);
      CHACHA20_QUARTERROUND(ctx->keystream32, 3, 4,  9, 14);
    }
  
  for (u32 i = 0; i < 16; i++)
    ctx->keystream32[i] += ctx->state[i];
  
  u32 *counter = ctx->state + 12;
  
  counter[0]++;

  if (counter[0] == 0) 
    {
      counter[1]++;
      
      if (counter[1] == 0)
	exit(1);
    }
}

void chacha20_init_context(chacha20_ctx_t *ctx, u8 key[], u8 nonce[], u64 counter)
{
  passman_memzero((u8 *)ctx, sizeof(chacha20_ctx_t));
  
  chacha20_init_block(ctx, key, nonce);
  chacha20_block_set_counter(ctx, counter);
  
  ctx->counter = counter;
  ctx->position = 64;
}

void chacha20_xor(chacha20_ctx_t *ctx, u8 *msg, u64 len, u8 *out)
{
  u8 *keystream8 = (u8*)ctx->keystream32;
  
  for (u64 i = 0; i < len; i++) 
    {
      if (ctx->position >= 64) 
	{
	  chacha20_block_next(ctx);
	  ctx->position = 0;
	}
      
      out[i] = msg[i] ^ keystream8[ctx->position];
      
      ctx->position++;
    }
}

void passman_chacha20(u8 *msg, u64 len, u8 *key, u8 *nonce, u8 *out) 
{
  chacha20_ctx_t ctx;
  
  chacha20_init_context(&ctx, key, nonce, 1);
  chacha20_xor(&ctx, msg, len, out);
}
