#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/random.h>

#include "types.h"
#include "const.h"
#include "error.h"
#include "memory.h"
#include "log.h"
#include "sha3.h"
#include "pbkdf2.h"
#include "chacha20.h"

void print_hash(u8 *hash, u64 len)
{
  for (u64 i = 0; i < len; i++)
    printf("%02x", hash[i]);
}

passman_log_t *read_log(ascii *fpath)
{
  FILE *fp = fopen(fpath, "rb");
  
  if (!fp)
    return NULL;
  else
    {
      passman_log_t *log = malloc(sizeof(passman_log_t));
      
      if (!log)
	{
	  fclose(fp);
	  return NULL;
	}
      else
	{
	  if (fread((u8 *)log, sizeof(passman_log_t), 1, fp) != 1)
	    {
	      free(log);
	      fclose(fp);
	      return NULL;
	    }
	  else
	    {
	      fclose(fp);
	      return log;
	    }
	}
    }
}

u8 write_log(ascii *fpath, passman_log_t *log)
{
  FILE *fp = fopen(fpath, "wb");
  
  if (!fp)
    return 0;
  else
    {
      if (fwrite((u8 *)log, sizeof(passman_log_t), 1, fp) != 1)
	{
	  fclose(fp);
	  return 0;
	}
      else
	{
	  fclose(fp);
	  
	  chmod(fpath, S_IRUSR | S_IWUSR);
	  
	  return 1;
	}
    }
}

void generate_log_key(u8 *password, u64 password_len, passman_log_t *log, u8 *key)
{
  passman_pbkdf2(passman_sha3_256_hmac, SHA3_256_HASH_LEN, password, password_len, log->salt, PASSMAN_SALT_LEN, log->iter, key, CHACHA20_KEY_LEN);
}

u8 lookup_free_log_entry(passman_log_t *log, passman_log_entry_t **curr, passman_log_entry_t **prev)
{
  for (u64 i = 0; i < PASSMAN_MAX_LOG_ENTRIES; i++)
    if (!log->entries[i].is_used)
      {
	(*curr) = &log->entries[i];
	(*prev) = (i) ? &log->entries[i - 1] : NULL;
	
	return 1;
      }
  
  return 0;
}

u8 generate_log_entry_hash(passman_log_entry_t *curr, passman_log_entry_t *prev, u8 *hash)
{
  u64 offset = 0;
  const u64 size = sizeof(passman_log_entry_t);
  u8 *buffer = malloc(sizeof(u8) * size);
  
  if (!buffer)
    return 0;
  else
    {
      passman_memcpy(buffer         , (u8 *)&curr->is_used, sizeof(curr->is_used)); offset += sizeof(curr->is_used);
      passman_memcpy(buffer + offset, (u8 *)&curr->id     , sizeof(curr->id));      offset += sizeof(curr->id);
      passman_memcpy(buffer + offset, (u8 *)curr->data    , sizeof(curr->data));    offset += sizeof(curr->data);
      
      if (prev)
	passman_memcpy(buffer + offset, prev->hash, sizeof(prev->hash)); 
      else
	passman_memzero(buffer + offset, SHA3_256_HASH_LEN);
      
      passman_sha3_256(buffer, size, hash);
      
      free(buffer);
      
      return 1;
    }
}

u8 generate_log_hmac(passman_log_t *log, u8 *log_key, u8 *hmac)
{
  u64 offset = 0;
  const u64 size = sizeof(passman_log_t) - SHA3_256_HASH_LEN;
  u8 *buffer = malloc(sizeof(u8) * size);

  if(!buffer)
    return 0;
  else
    {
      passman_memzero(buffer, size);
  
      passman_memcpy(buffer         , (u8 *)log->entries      , sizeof(log->entries));       offset += sizeof(log->entries);
      passman_memcpy(buffer + offset, (u8 *)&log->iter        , sizeof(log->iter));          offset += sizeof(log->iter);
      passman_memcpy(buffer + offset, log->salt               , sizeof(log->salt));          offset += sizeof(log->salt);
      passman_memcpy(buffer + offset, log->nonce              , sizeof(log->nonce));         offset += sizeof(log->nonce);
      passman_memcpy(buffer + offset, log->entries_hmac       , sizeof(log->entries_hmac));  
      
      passman_sha3_256_hmac(buffer, size, log_key, CHACHA20_KEY_LEN, hmac);
      
      free(buffer);
      
      return 1;
    }
}

u8 add_log_entry(passman_log_t *log, ascii *data, u32 id)
{
  time_t i_time = time(NULL);
  struct tm i_tm = *localtime(&i_time);
  passman_log_entry_t *curr = NULL, *prev = NULL;
  
  if (lookup_free_log_entry(log, &curr, &prev))
    {
      curr->is_used = 1;
      
      getrandom(&curr->id, sizeof(curr->id), 0);
      
      if (id)
	snprintf(curr->data, sizeof(curr->data) - 1, "%d-%02d-%02d; %02d:%02d:%02d; %s; %u",
		 i_tm.tm_year + 1900,
		 i_tm.tm_mon  + 1,
		 i_tm.tm_mday,
		 i_tm.tm_hour,
		 i_tm.tm_min,
		 i_tm.tm_sec,
		 data,
		 id);
      else
	snprintf(curr->data, sizeof(curr->data) - 1, "%d-%02d-%02d; %02d:%02d:%02d; %s",
		 i_tm.tm_year + 1900,
		 i_tm.tm_mon  + 1,
		 i_tm.tm_mday,
		 i_tm.tm_hour,
		 i_tm.tm_min,
		 i_tm.tm_sec,
		 data);
      
      generate_log_entry_hash(curr, prev, curr->hash);
      
      return 1;
    }
  else
    return 0;
}

u8 check_log(passman_log_t *log)
{
  u8 hash[SHA3_256_HASH_LEN];
  
  for (u64 i = 1; i < PASSMAN_MAX_LOG_ENTRIES; i++)
    if(log->entries[i].is_used)
      {
	generate_log_entry_hash(&log->entries[i], &log->entries[i - 1], hash);
	
	if (passman_memcmp(log->entries[i].hash, hash, SHA3_256_HASH_LEN))
	  return 0;
      }
  
  return 1;
}

void print_log(passman_log_t *log)
{
  for (u64 i = 0; i < PASSMAN_MAX_LOG_ENTRIES; i++)
    if (log->entries[i].is_used)
      {
	printf("%-10u; %-60s; ", log->entries[i].id, log->entries[i].data);
	print_hash(log->entries[i].hash, SHA3_256_HASH_LEN);
	putchar('\n');
      }
}
