/*
  
  If you can't understand the code, go flip burgers at McDonald's.
  Sincerely, yaspr (2024)
  
*/
#include <pwd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <termios.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/random.h>
#include <sys/resource.h>

#include "types.h"
#include "const.h"
#include "error.h"
#include "memory.h"
#include "log.h"
#include "sha3.h"
#include "pbkdf2.h"
#include "chacha20.h"
#include "passman.h"

u64 get_random_ranged_number(u64 min, u64 max)
{
  u64 r = 0;
  
  getrandom(&r, sizeof(r), 0);
  
  return r % (max - min + 1) + min;
}

u8 get_random_buffer(u8 *buffer, u64 n)
{
  return ((u64)getrandom(buffer, n, 0) == n);
}

void get_secret(ascii secret[PASSMAN_MAX_STR], u64 *len, u8 mode)
{
  u64 i = 0;
  ascii c = 0;
  
  while ((i < PASSMAN_MAX_STR - 1) && (c != '\n'))
    {
      struct termios t1, t2;
      
      tcgetattr(STDIN_FILENO, &t1);
      t2 = t1;
      t2.c_lflag &= ~ICANON;
      
      if (mode == PASSMAN_NOECHO)
	t2.c_lflag &= ~ECHO;
      else
	t2.c_lflag |= ECHO;
      
      tcsetattr(STDIN_FILENO, TCSANOW, &t2);
      
      c = getchar();
      
      if (c != '\n')
	secret[i++] = c;
      
      tcsetattr(STDIN_FILENO, TCSANOW, &t1);
    }
  
  secret[i] = 0;
  (*len) = i;
  
  if (c != '\n' || mode == PASSMAN_NOECHO)
    putchar('\n');
}

void generate_db_key(u8 *password, u64 password_len, passman_db_t *db, u8 *key)
{
  passman_pbkdf2(passman_sha3_256_hmac, SHA3_256_HASH_LEN, password, password_len, db->salt, PASSMAN_SALT_LEN, db->iter, key, CHACHA20_KEY_LEN);
}

void generate_entry_key(u8 *password, u64 password_len, passman_db_entry_t *entry, u8 *key)
{
  passman_pbkdf2(passman_sha3_256_hmac, SHA3_256_HASH_LEN, password, password_len, entry->salt, PASSMAN_SALT_LEN, entry->iter, key, CHACHA20_KEY_LEN);
}

u8 check_db_directory(ascii *user_home_dir)
{
  struct stat sb;
  ascii passman_path[PASSMAN_MAX_PATH];
  
  snprintf(passman_path, sizeof(passman_path), "%s/.passman", user_home_dir);
  
  return !(stat(passman_path, &sb) < 0);
}

u8 make_db_directory(ascii *user_home_dir)
{
  struct stat sb;
  ascii passman_path[PASSMAN_MAX_PATH];
  
  snprintf(passman_path, sizeof(passman_path), "%s/.passman", user_home_dir);
  
  return (stat(passman_path, &sb) < 0) && !mkdir(passman_path, 0700);
}

passman_db_t *read_db(ascii *fpath)
{
  FILE *fp = fopen(fpath, "rb");
  
  if (!fp)
    return NULL;
  else
    {
      passman_db_t *db = malloc(sizeof(passman_db_t));
      
      if (!db)
	{
	  fclose(fp);
	  return NULL;
	}
      else
	{
	  if (fread((u8 *)db, sizeof(passman_db_t), 1, fp) != 1)
	    {
	      free(db);
	      fclose(fp);
	      return NULL;
	    }
	  else
	    {  
	      fclose(fp);
	      return db;
	    }
	}
    }
}

u8 write_db(ascii *fpath, passman_db_t *db)
{
  FILE *fp = fopen(fpath, "wb");
  
  if (!fp)
    return 0;
  else
    {
      if (fwrite((u8 *)db, sizeof(passman_db_t), 1, fp) != 1)
	{
	  fclose(fp);
	  return 0;
	}
      else
	{
	  fclose(fp);
	  
	  chmod(fpath, S_IRUSR | S_IWUSR);
	  
	  return 1;
	}
    }
}

passman_db_entry_t *lookup_free_db_entry(passman_db_t *db)
{
  for (u64 i = 0; i < PASSMAN_MAX_ENTRIES; i++)
    if (!db->entries[i].is_used)
      return &db->entries[i];
  
  return NULL;
}

u8 generate_db_entry_hmac(passman_db_entry_t *entry, u8 *db_key, u8 *hmac)
{
  u64 offset = 0;
  const u64 size = sizeof(passman_db_entry_t) - SHA3_256_HASH_LEN;
  u8 *buffer = malloc(sizeof(u8) * size);
  
  if (!buffer)
    return 0;
  else
    {
      passman_memzero(buffer, size);
      
      passman_memcpy(buffer         , (u8 *)&entry->is_used     , sizeof(entry->is_used));     offset += sizeof(entry->is_used);
      passman_memcpy(buffer + offset, (u8 *)&entry->id          , sizeof(entry->id));          offset += sizeof(entry->id);
      passman_memcpy(buffer + offset, (u8 *)entry->description  , sizeof(entry->description)); offset += sizeof(entry->description);
      passman_memcpy(buffer + offset, (u8 *)entry->login        , sizeof(entry->login));       offset += sizeof(entry->login);
      passman_memcpy(buffer + offset, (u8 *)entry->password     , sizeof(entry->password));    offset += sizeof(entry->password);
      passman_memcpy(buffer + offset, (u8 *)&entry->iter        , sizeof(entry->iter));        offset += sizeof(entry->iter);
      passman_memcpy(buffer + offset, entry->salt               , sizeof(entry->salt));        offset += sizeof(entry->salt);
      passman_memcpy(buffer + offset, entry->nonce              , sizeof(entry->nonce));       
      
      passman_sha3_256_hmac(buffer, size, db_key, CHACHA20_KEY_LEN, hmac);
      
      free(buffer);
      
      return 1;
    }
}

u8 generate_db_hmac(passman_db_t *db, u8 *db_key, u8 *hmac)
{
  u64 offset = 0;
  const u64 size = sizeof(passman_db_t) - SHA3_256_HASH_LEN;
  u8 *buffer = malloc(sizeof(u8) * size);
  
  if (!buffer)
    return 0;
  else
    {
      passman_memzero(buffer, size);
      
      passman_memcpy(buffer         , (u8 *)db->entries      , sizeof(db->entries));       offset += sizeof(db->entries);
      passman_memcpy(buffer + offset, (u8 *)&db->iter        , sizeof(db->iter));          offset += sizeof(db->iter);
      passman_memcpy(buffer + offset, db->salt               , sizeof(db->salt));          offset += sizeof(db->salt);
      passman_memcpy(buffer + offset, db->nonce              , sizeof(db->nonce));         offset += sizeof(db->nonce);
      passman_memcpy(buffer + offset, db->entries_hmac       , sizeof(db->entries_hmac));  
      
      passman_sha3_256_hmac(buffer, size, db_key, CHACHA20_KEY_LEN, hmac);
      
      free(buffer);
      
      return 1;  
    }
}

u8 add_db_entry(passman_db_t *db, u8 *db_key)
{
  u64 login_len = 0;
  u64 password_len = 0;
  u64 description_len = 0;
  u8 entry_key[CHACHA20_KEY_LEN];
  passman_db_entry_t *entry = lookup_free_db_entry(db);
  
  if (!entry)
    return 0;
  
  entry->is_used = 1;
  
  if (!get_random_buffer((u8 *)&entry->id, sizeof(u32)))
    return 0;
  
  printf("description: ");
  get_secret(entry->description, &description_len, PASSMAN_ECHO);
  
  printf("login      : ");
  get_secret(entry->login, &login_len, PASSMAN_ECHO);
  
  printf("password   : ");
  get_secret(entry->password, &password_len, PASSMAN_ECHO);
  
  entry->iter = get_random_ranged_number(PASSMAN_ITER_MIN, PASSMAN_ITER_MAX);
  
  if (!get_random_buffer(entry->salt, sizeof(entry->salt)))
    return 0;
  
  if (!get_random_buffer(entry->nonce, sizeof(entry->nonce)))
    return 0;
  
  generate_entry_key(db_key, CHACHA20_KEY_LEN, entry, entry_key);
  
  passman_chacha20((u8 *)entry->login, sizeof(entry->login), entry_key, entry->nonce, (u8 *)entry->login);
  passman_chacha20((u8 *)entry->password, sizeof(entry->password), entry_key, entry->nonce, (u8 *)entry->password);
  
  if (generate_db_entry_hmac(entry, db_key, entry->hmac))
    {
      passman_memzero(entry_key, sizeof(entry_key));
      return 1;
    }
  else
    return 0;
}

u8 del_db_entry(passman_db_t *db, u32 id)
{
  for (u64 i = 0; i < PASSMAN_MAX_ENTRIES; i++)
    if (db->entries[i].is_used)
      if (db->entries[i].id == id)
	{
	  db->entries[i].is_used = 0;
	  
	  db->entries[i].id = 0;
	  passman_memzero((u8 *)db->entries[i].description, sizeof(db->entries[i].description));
	  passman_memzero((u8 *)db->entries[i].login, sizeof(db->entries[i].login));
	  passman_memzero((u8 *)db->entries[i].password, sizeof(db->entries[i].password));
	  
	  db->entries[i].iter = 0;
	  passman_memzero((u8 *)db->entries[i].salt, sizeof(db->entries[i].salt));
	  passman_memzero((u8 *)db->entries[i].nonce, sizeof(db->entries[i].nonce));
	  passman_memzero((u8 *)db->entries[i].hmac, sizeof(db->entries[i].hmac));
	  
	  return 1;
	}
  
  return 0;
}

void print_db(passman_db_t *db)
{
  u64 n_entries = 0;
  
  for (u64 i = 0; i < PASSMAN_MAX_ENTRIES; i++)
    if (db->entries[i].is_used)
      {
	passman_db_entry_t *entry = &db->entries[i];
	
	printf("\n# entry %llu #\n", n_entries + 1);
	printf(" id         : %u\n", entry->id);
	printf(" description: %s\n", entry->description);
	
	n_entries++;
      }
  
  if (n_entries)
    printf("\ntotal number of entries: %llu\n", n_entries);
  else
    printf("\nno entries found\n");
}

void print_db_entry(passman_db_t *db, u8 *db_key, u32 id)
{
  for (u64 i = 0; i < PASSMAN_MAX_ENTRIES; i++)
    if (db->entries[i].is_used)
      if (db->entries[i].id == id)
	{
	  u8 entry_key[CHACHA20_KEY_LEN];
	  passman_db_entry_t *entry = &db->entries[i];
	  
	  printf("\n# entry #\n");
	  printf(" id         : %u\n", entry->id);
	  printf(" description: %s\n", entry->description); 
	  
	  generate_entry_key(db_key, CHACHA20_KEY_LEN, entry, entry_key);
	  
	  passman_chacha20((u8 *)entry->login, sizeof(entry->login), entry_key, entry->nonce, (u8 *)entry->login);
	  passman_chacha20((u8 *)entry->password, sizeof(entry->password), entry_key, entry->nonce, (u8 *)entry->password);
	  
	  printf(" login      : %s\n", entry->login);
	  printf(" password   : %s\n", entry->password);
	  
	  passman_chacha20((u8 *)entry->login, sizeof(entry->login), entry_key, entry->nonce, (u8 *)entry->login);
	  passman_chacha20((u8 *)entry->password, sizeof(entry->password), entry_key, entry->nonce, (u8 *)entry->password);
	  
	  passman_memzero(entry_key, sizeof(entry_key));
	}
}

void print_help(ascii *binary)
{
  const ascii logo[] =
    "┌─┐┌─┐┌─┐┌─┐┌┬┐┌─┐┌┐┌\n"
    "├─┘├─┤└─┐└─┐│││├─┤│││\n"
    "┴  ┴ ┴└─┘└─┘┴ ┴┴ ┴┘└┘\n";
  
  printf("usage: %s [command]\n", binary);
  printf("\n%s", logo);
  printf("password manager v%s - %s - %s\n\n", PASSMAN_VERSION, PASSMAN_LICENSE, PASSMAN_YEAR);
  
  printf("commands:\n");
  printf("    init  : initialize the database\n");
  printf("    add   : add a new entry to the database\n");
  printf("    del id: delete an entry from the database\n");
  printf("    get id: show the content of an entry\n");
  printf("    list  : show the list of entries\n");
  printf("    log   : show the log of activities\n");
  printf("    help  : show this help\n");
  printf("\ncontact: yaspr@pm.me\n");
}

int main(int argc, char **argv)
{
  struct rlimit limit;
  limit.rlim_cur = 0;
  limit.rlim_max = 0;
  
  if (setrlimit(RLIMIT_CORE, &limit) != 0)
    passman_error("passman", 1, "cannot set rlimit!\n");
    
  if (mlockall(MCL_CURRENT | MCL_FUTURE))
    passman_error("passman", 1, "cannot lock memory!\n");
  
  if (argc < 2)
    return print_help(argv[0]), 1;
  
  u32 id = 0;
  u8 *command = (u8 *)argv[1];
  u8 mode = PASSMAN_MODE_NONE;
  u64 len = strnlen(argv[1], PASSMAN_MAX_STR);
  
  if (len == 4 && !passman_memcmp(command, (u8 *)"init", 4))
    mode = PASSMAN_MODE_INIT;
  else
    if (len == 3 && !passman_memcmp(command, (u8 *)"add", 3))
      mode = PASSMAN_MODE_ADD;
    else
      if (len == 3 && !passman_memcmp(command, (u8 *)"del", 3))
	{
	  mode = PASSMAN_MODE_DEL;
	  
	  if (argc < 3)
	    passman_error("passman", 1, "missing 'id' argument\n");
	  
	  id = (u32)atoi(argv[2]);
	}
      else
	if (len == 3 && !passman_memcmp(command, (u8 *)"get", 3))
	  {
	    mode = PASSMAN_MODE_GET;
	    
	    if (argc < 3)
	      passman_error("passman", 1, "missing 'id' argument\n");
	    else
	      id = (u32)atoi(argv[2]);
	  }
	else
	  if (len == 4 && !passman_memcmp(command, (u8 *)"list", 4))
	    mode = PASSMAN_MODE_LIST;
	  else
	    if (len == 3 && !passman_memcmp(command, (u8 *)"log", 3))
	      mode = PASSMAN_MODE_LOG;
	    else
	      if (len == 4 && !passman_memcmp(command, (u8 *)"help", 4))
		return print_help(argv[0]), 1;
	      else
		passman_error("passman", 1, "unknown command '%s', try: %s help\n", command, argv[0]);
  
  ascii db_file_path[PASSMAN_MAX_PATH];
  ascii log_file_path[PASSMAN_MAX_PATH];
  struct passwd *user_passwd = getpwuid(getuid());
  
  snprintf(db_file_path, sizeof(db_file_path), "%s/.passman/db", user_passwd->pw_dir);
  snprintf(log_file_path, sizeof(log_file_path), "%s/.passman/log", user_passwd->pw_dir);
  
  if (mode == PASSMAN_MODE_INIT)
    {
      if (!check_db_directory(user_passwd->pw_dir))
	{
	  passman_db_t db;
	  passman_log_t log;
	  u64 password1_len = 0;
	  u64 password2_len = 0;
	  u8 db_key[CHACHA20_KEY_LEN];
	  u8 log_key[CHACHA20_KEY_LEN];
	  ascii password1[PASSMAN_MAX_STR];
	  ascii password2[PASSMAN_MAX_STR];
	  
	  db.iter = get_random_ranged_number(PASSMAN_ITER_MIN, PASSMAN_ITER_MAX);

	  if (!get_random_buffer(db.salt, sizeof(db.salt)))
	    passman_error("init", 1, "getrandom failure\n");
	  
	  if (!get_random_buffer(db.nonce, sizeof(db.nonce)))
	    passman_error("init", 1, "getrandom failure\n");
	  
	  log.iter = get_random_ranged_number(PASSMAN_ITER_MIN, PASSMAN_ITER_MAX);

	  if (!get_random_buffer(log.salt, sizeof(log.salt)))
	    passman_error("init", 1, "getrandom failure\n");

	  if (!get_random_buffer(log.nonce, sizeof(log.nonce)))
	    passman_error("init", 1, "getrandom failure\n");
	  
	  passman_memzero((u8 *)db.entries, sizeof(db.entries));
	  passman_memzero((u8 *)log.entries, sizeof(log.entries));
	  
	  printf("enter master password      : ");
	  get_secret(password1, &password1_len, PASSMAN_NOECHO);
	  
	  printf("enter master password again: ");
	  get_secret(password2, &password2_len, PASSMAN_NOECHO);
	  
	  if ((password1_len != password2_len) || passman_memcmp((u8 *)password1, (u8 *)password2, password1_len))
	    passman_error("init", 1, "input mismatch!\n");
	  
	  passman_memzero((u8 *)password2, sizeof(password2)); password2_len = 0;
	  
	  generate_db_key((u8 *)password1, password1_len, &db, db_key);
	  generate_log_key((u8 *)password1, password1_len, &log, log_key);
	  
	  passman_memzero((u8 *)password1, sizeof(password1)); password1_len = 0;
	  
	  add_log_entry(&log, "init", 0);
	  
	  passman_sha3_256_hmac((u8 *)db.entries, sizeof(db.entries), db_key, sizeof(db_key), db.entries_hmac);
	  passman_chacha20((u8 *)db.entries, sizeof(db.entries), db_key, db.nonce, (u8 *)db.entries);
	  
	  passman_sha3_256_hmac((u8 *)log.entries, sizeof(log.entries), log_key, sizeof(log_key), log.entries_hmac);
	  passman_chacha20((u8 *)log.entries, sizeof(log.entries), log_key, log.nonce, (u8 *)log.entries);
	  
	  if (!generate_db_hmac(&db, db_key, db.hmac))
	    passman_error("init", 1, "cannot compute database HMAC\n");
	  
	  passman_memzero(db_key, sizeof(db_key));
	  
	  if (!generate_log_hmac(&log, log_key, log.hmac))
	    passman_error("init", 1, "cannot compute database HMAC\n");
	  
	  passman_memzero(log_key, sizeof(log_key));
	  
	  if (make_db_directory(user_passwd->pw_dir))
	    {
	      if (!write_db(db_file_path, &db) || !write_log(log_file_path, &log))
		passman_error("init", 1, "cannot write into database or log\n");
	      else
		{
		  passman_memzero((u8 *)&db, sizeof(db));
		  passman_memzero((u8 *)&log, sizeof(log));
		  
		  return 0;
		}
	    }
	  else
	    passman_error("passman", 1, "cannot create '%s/.passman' directory\n", user_passwd->pw_dir); 
	}
      else
	passman_error("init", 1, "'%s/.passman' directory already exists\n", user_passwd->pw_dir); 
    }
  
  if (mode == PASSMAN_MODE_ADD)
    {
      u64 password_len = 0;
      u8 db_key[CHACHA20_KEY_LEN];
      u8 log_key[CHACHA20_KEY_LEN];
      u8 db_hmac[SHA3_256_HASH_LEN];
      u8 log_hmac[SHA3_256_HASH_LEN];
      ascii password[PASSMAN_MAX_STR];
      passman_db_t *db = read_db(db_file_path);
      
      if (!db)
	passman_error("add", 1, "cannot read database\n");
      else
	{
	  passman_log_t *log = read_log(log_file_path);
	  
	  if (!log)
	    {
	      free(db);
	      passman_error("add", 1, "cannot read log\n");
	    }
	  else
	    {
	      printf("master password: ");
	      get_secret(password, &password_len, PASSMAN_NOECHO);
	      
	      generate_db_key((u8 *)password, password_len, db, db_key);
	      generate_log_key((u8 *)password, password_len, log, log_key);
	  
	      passman_memzero((u8 *)password, sizeof(password)); password_len = 0;
	  
	      if (!generate_db_hmac(db, db_key, db_hmac))
		passman_error("add", 1, "cannot compute database HMAC\n");
	      
	      if (!generate_log_hmac(log, log_key, log_hmac))
		passman_error("add", 1, "cannot compute log HMAC\n");
		
	      if (!passman_memcmp(db_hmac, db->hmac, SHA3_256_HASH_LEN) && !passman_memcmp(log_hmac, log->hmac, SHA3_256_HASH_LEN))
		{
		  passman_chacha20((u8 *)db->entries, sizeof(db->entries), db_key, db->nonce, (u8 *)db->entries);      
		  passman_sha3_256_hmac((u8 *)db->entries, sizeof(db->entries), db_key, sizeof(db_key), db_hmac);
		  
		  passman_chacha20((u8 *)log->entries, sizeof(log->entries), log_key, log->nonce, (u8 *)log->entries);
		  passman_sha3_256_hmac((u8 *)log->entries, sizeof(log->entries), log_key, CHACHA20_KEY_LEN, log_hmac);
		  
		  if (!passman_memcmp(db_hmac, db->entries_hmac, SHA3_256_HASH_LEN) && !passman_memcmp(log_hmac, log->entries_hmac, SHA3_256_HASH_LEN))
		    if (add_db_entry(db, db_key) && add_log_entry(log, "add", 0))
		      {
			if (!get_random_buffer(db->nonce, sizeof(db->nonce)))
			  passman_error("init", 1, "getrandom failure\n");
			
			if (!get_random_buffer(log->nonce, sizeof(log->nonce)))
			  passman_error("init", 1, "getrandom failure\n");
			
			passman_sha3_256_hmac((u8 *)db->entries, sizeof(db->entries), db_key, sizeof(db_key), db->entries_hmac);
			passman_chacha20((u8 *)db->entries, sizeof(db->entries), db_key, db->nonce, (u8 *)db->entries);      
			
			passman_sha3_256_hmac((u8 *)log->entries, sizeof(log->entries), log_key, sizeof(log_key), log->entries_hmac);
			passman_chacha20((u8 *)log->entries, sizeof(log->entries), log_key, log->nonce, (u8 *)log->entries);      
			
			if (!generate_db_hmac(db, db_key, db->hmac))
			  passman_error("add", 1, "cannot compute database HMAC\n");

			if (!generate_log_hmac(log, log_key, log->hmac))
			  passman_error("add", 1, "cannot compute log HMAC\n");
					
			passman_memzero(db_key, sizeof(db_key));
			passman_memzero(log_key, sizeof(log_key));
			
			if (!write_db(db_file_path, db) || !write_log(log_file_path, log))
			  {
			    passman_memzero((u8 *)db, sizeof(passman_db_t));
			    passman_memzero((u8 *)log, sizeof(passman_log_t));
			    free(db);
			    free(log);
			    
			    passman_error("add", 1, "cannot write into database or log\n");
			  }
			else
			  {
			    passman_memzero((u8 *)db, sizeof(passman_db_t));
			    passman_memzero((u8 *)log, sizeof(passman_log_t));
			    free(db);
			    free(log);
			    
			    return 0;
			  }
		      }
		    else
		      {
			passman_memzero((u8 *)db, sizeof(passman_db_t));
			passman_memzero((u8 *)log, sizeof(passman_log_t));
			free(db);
			free(log);
			
			passman_error("add", 1, "cannot add entry to database or write to log\n");
		      }
		  else
		    {
		      passman_memzero((u8 *)db, sizeof(passman_db_t));
		      passman_memzero((u8 *)log, sizeof(passman_log_t));
		      free(db);
		      free(log);
		      
		      passman_error("add", 1, "invalid database or log signature\n");
		    }
		}
	      else
		{
		  passman_memzero((u8 *)db, sizeof(passman_db_t));
		  passman_memzero((u8 *)log, sizeof(passman_log_t));
		  free(db);
		  free(log);
		  
		  passman_error("add", 1, "invalid master password\n");
		}
	    }
	}
    }
  
  if (mode == PASSMAN_MODE_DEL)
    {
      u64 password_len = 0;
      u8 db_key[CHACHA20_KEY_LEN];
      u8 log_key[CHACHA20_KEY_LEN];
      u8 db_hmac[SHA3_256_HASH_LEN];
      u8 log_hmac[SHA3_256_HASH_LEN];
      ascii password[PASSMAN_MAX_STR];
      passman_db_t *db = read_db(db_file_path);
      
      if (!db)
	passman_error("del", 1, "cannot read database\n");
      else
	{
	  passman_log_t *log = read_log(log_file_path);
	  
	  if (!log)
	    {
	      free(db);
	      passman_error("del", 1, "cannot read log\n");
	    }
	  else
	    {
	      printf("master password: ");
	      get_secret(password, &password_len, PASSMAN_NOECHO);
      
	      generate_db_key((u8 *)password, password_len, db, db_key);
	      generate_log_key((u8 *)password, password_len, log, log_key);
      
	      passman_memzero((u8 *)password, sizeof(password)); password_len = 0;
	      
	      if (!generate_db_hmac(db, db_key, db_hmac))
		passman_error("del", 1, "cannot compute database HMAC\n");
	      
	      if (!generate_log_hmac(log, log_key, log_hmac))
		passman_error("del", 1, "cannot compute log HMAC\n");
	      
	      if (!passman_memcmp(db_hmac, db->hmac, SHA3_256_HASH_LEN) && !passman_memcmp(log_hmac, log->hmac, SHA3_256_HASH_LEN))
		{
		  passman_chacha20((u8 *)db->entries, sizeof(db->entries), db_key, db->nonce, (u8 *)db->entries);      
		  passman_sha3_256_hmac((u8 *)db->entries, sizeof(db->entries), db_key, sizeof(db_key), db_hmac);
		  
		  passman_chacha20((u8 *)log->entries, sizeof(log->entries), log_key, log->nonce, (u8 *)log->entries);
		  passman_sha3_256_hmac((u8 *)log->entries, sizeof(log->entries), log_key, CHACHA20_KEY_LEN, log_hmac);
	  
		  if (!passman_memcmp(db_hmac, db->entries_hmac, SHA3_256_HASH_LEN) && !passman_memcmp(log_hmac, log->entries_hmac, SHA3_256_HASH_LEN))
		    if (del_db_entry(db, id) && add_log_entry(log, "del", id))
		      {
			if (!get_random_buffer(db->nonce, sizeof(db->nonce)))
			  passman_error("del", 1, "getrandom failure\n");

			if (!get_random_buffer(log->nonce, sizeof(log->nonce)))
			  passman_error("del", 1, "getrandom failure\n");
		
			passman_sha3_256_hmac((u8 *)db->entries, sizeof(db->entries), db_key, sizeof(db_key), db->entries_hmac);
			passman_chacha20((u8 *)db->entries, sizeof(db->entries), db_key, db->nonce, (u8 *)db->entries);

			passman_sha3_256_hmac((u8 *)log->entries, sizeof(log->entries), log_key, sizeof(log_key), log->entries_hmac);
			passman_chacha20((u8 *)log->entries, sizeof(log->entries), log_key, log->nonce, (u8 *)log->entries);
		
			if (!generate_db_hmac(db, db_key, db->hmac))
			  passman_error("del", 1, "cannot compute database HMAC\n");

			if (!generate_log_hmac(log, log_key, log->hmac))
			  passman_error("del", 1, "cannot compute log HMAC\n");
		
			passman_memzero(db_key, sizeof(db_key));
			passman_memzero(log_key, sizeof(log_key));
		
			if (!write_db(db_file_path, db) || !write_log(log_file_path, log))
			  {
			    passman_memzero((u8 *)db, sizeof(passman_db_t));
			    passman_memzero((u8 *)log, sizeof(passman_log_t));
			    free(db);
			    free(log);
		    
			    passman_error("del", 1, "cannot write into database or log\n");
			  }
			else
			  {
			    passman_memzero((u8 *)db, sizeof(passman_db_t));
			    passman_memzero((u8 *)log, sizeof(passman_log_t));
			    free(db);
			    free(log);

			    return 0;
			  }
		      }
		    else
		      {
			passman_memzero((u8 *)db, sizeof(passman_db_t));
			passman_memzero((u8 *)log, sizeof(passman_log_t));
			free(db);
			free(log);

			passman_error("del", 1, "cannot delete entry from database or write to log\n");
		      }
		  else
		    {
		      passman_memzero((u8 *)db, sizeof(passman_db_t));
		      passman_memzero((u8 *)log, sizeof(passman_log_t));
		      free(db);
		      free(log);
	      
		      passman_error("del", 1, "invalid database or log signature\n");
		    }
		}
	      else
		{
		  passman_memzero((u8 *)db, sizeof(passman_db_t));
		  passman_memzero((u8 *)log, sizeof(passman_log_t));
		  free(db);
		  free(log);
	  
		  passman_error("del", 1, "invalid master password\n");
		}
	    }
	}
    }
  
  if (mode == PASSMAN_MODE_GET)
    {
      u64 password_len = 0;
      u8 db_key[CHACHA20_KEY_LEN];
      u8 log_key[CHACHA20_KEY_LEN];
      u8 db_hmac[SHA3_256_HASH_LEN];
      u8 log_hmac[SHA3_256_HASH_LEN];
      ascii password[PASSMAN_MAX_STR];
      passman_db_t *db = read_db(db_file_path);
      
      if (!db)
	passman_error("get", 1, "cannot read database\n");
      else
	{
	  passman_log_t *log = read_log(log_file_path);
	  
	  if (!log)
	    {
	      free(db);
	      passman_error("get", 1, "cannot read log\n");
	    }
	  else
	    {
	      printf("master password: ");
	      get_secret(password, &password_len, PASSMAN_NOECHO);
	      
	      generate_db_key((u8 *)password, password_len, db, db_key);
	      generate_log_key((u8 *)password, password_len, log, log_key);
      
	      passman_memzero((u8 *)password, sizeof(password)); password_len = 0;
      
	      if (!generate_db_hmac(db, db_key, db_hmac))
		passman_error("del", 1, "cannot compute database HMAC\n");
	      
	      if (!generate_log_hmac(log, log_key, log_hmac))
		passman_error("del", 1, "cannot compute log HMAC\n");
      
	      if (!passman_memcmp(db_hmac, db->hmac, SHA3_256_HASH_LEN) && !passman_memcmp(log_hmac, log->hmac, SHA3_256_HASH_LEN))
		{
		  passman_chacha20((u8 *)db->entries, sizeof(db->entries), db_key, db->nonce, (u8 *)db->entries);
		  passman_sha3_256_hmac((u8 *)db->entries, sizeof(db->entries), db_key, sizeof(db_key), db_hmac);

		  passman_chacha20((u8 *)log->entries, sizeof(log->entries), log_key, log->nonce, (u8 *)log->entries);
		  passman_sha3_256_hmac((u8 *)log->entries, sizeof(log->entries), log_key, sizeof(log_key), log_hmac);
	  
		  if (!passman_memcmp(db_hmac, db->entries_hmac, SHA3_256_HASH_LEN) && !passman_memcmp(log_hmac, log->entries_hmac, SHA3_256_HASH_LEN))
		    {
		      print_db_entry(db, db_key, id);
	      
		      passman_memzero(db_key, sizeof(db_key));
	      
		      if (add_log_entry(log, "get", id))
			{
			  if (!get_random_buffer(log->nonce, sizeof(log->nonce)))
			    passman_error("get", 1, "getrandom failure\n");
			  
			  passman_sha3_256_hmac((u8 *)log->entries, sizeof(log->entries), log_key, sizeof(log_key), log->entries_hmac);
			  passman_chacha20((u8 *)log->entries, sizeof(log->entries), log_key, log->nonce, (u8 *)log->entries);
			  
			  if (!generate_log_hmac(log, log_key, log->hmac))
			    passman_error("get", 1, "cannot compute log HMAC\n");
			  
			  passman_memzero(log_key, sizeof(log_key));
			  
			  if (!write_log(log_file_path, log))
			    {
			      passman_memzero((u8 *)db, sizeof(passman_db_t));
			      passman_memzero((u8 *)log, sizeof(passman_log_t));
			      free(db);
			      free(log);

			      passman_error("get", 1, "cannot write into log\n");
			    }
			  else
			    {
			      passman_memzero((u8 *)db, sizeof(passman_db_t));
			      passman_memzero((u8 *)log, sizeof(passman_log_t));
			      free(db);
			      free(log);
		      
			      return 0;
			    }
			}
		      else
			{
			  passman_memzero((u8 *)db, sizeof(passman_db_t));
			  passman_memzero((u8 *)log, sizeof(passman_log_t));
			  free(db);
			  free(log);

			  passman_error("get", 1, "cannot add entry or write to log\n");
			}
		    }
		  else
		    {
		      passman_memzero((u8 *)db, sizeof(passman_db_t));
		      passman_memzero((u8 *)log, sizeof(passman_log_t));
		      free(db);
		      free(log);
	      
		      passman_error("get", 1, "invalid database or log signature\n");
		    }
		}
	      else
		{
		  passman_memzero((u8 *)db, sizeof(passman_db_t));
		  passman_memzero((u8 *)log, sizeof(passman_log_t));
		  free(db);
		  free(log);
	  
		  passman_error("get", 1, "invalid master password\n");
		}
	    }
	}
    }
  
  if (mode == PASSMAN_MODE_LIST)
    {
      u64 password_len = 0;
      u8 db_key[CHACHA20_KEY_LEN];
      u8 log_key[CHACHA20_KEY_LEN];
      u8 db_hmac[SHA3_256_HASH_LEN];
      u8 log_hmac[SHA3_256_HASH_LEN];
      ascii password[PASSMAN_MAX_STR];
      passman_db_t *db = read_db(db_file_path);
      
      if (!db)
	passman_error("list", 1, "cannot read database\n");
      else
	{
	  passman_log_t *log = read_log(log_file_path);
	  
	  if (!log)
	    {
	      free(db);
	      passman_error("list", 1, "cannot read log\n");
	    }
	  else
	    {
	      printf("master password: ");
	      get_secret(password, &password_len, PASSMAN_NOECHO);
      
	      generate_db_key((u8 *)password, password_len, db, db_key);
	      generate_log_key((u8 *)password, password_len, log, log_key);
      
	      passman_memzero((u8 *)password, sizeof(password)); password_len = 0;
      
	      if (!generate_db_hmac(db, db_key, db_hmac))
		passman_error("del", 1, "cannot compute database HMAC\n");
	      
	      if (!generate_log_hmac(log, log_key, log_hmac))
		passman_error("del", 1, "cannot compute log HMAC\n");
	      
	      if (!passman_memcmp(db_hmac, db->hmac, SHA3_256_HASH_LEN) && !passman_memcmp(log_hmac, log->hmac, SHA3_256_HASH_LEN))
		{
		  passman_chacha20((u8 *)db->entries, sizeof(db->entries), db_key, db->nonce, (u8 *)db->entries);
		  passman_sha3_256_hmac((u8 *)db->entries, sizeof(db->entries), db_key, sizeof(db_key), db_hmac);

		  passman_chacha20((u8 *)log->entries, sizeof(log->entries), log_key, log->nonce, (u8 *)log->entries);
		  passman_sha3_256_hmac((u8 *)log->entries, sizeof(log->entries), log_key, sizeof(log_key), log_hmac);
	  
		  if (!passman_memcmp(db_hmac, db->entries_hmac, SHA3_256_HASH_LEN) && !passman_memcmp(log_hmac, log->entries_hmac, SHA3_256_HASH_LEN))
		    {
		      print_db(db);
	      
		      passman_memzero(db_key, sizeof(db_key));
	      
		      if (add_log_entry(log, "list", 0))
			{
			  if (!get_random_buffer(log->nonce, sizeof(log->nonce)))
			    passman_error("list", 1, "getrandom failure\n");
			  
			  passman_sha3_256_hmac((u8 *)log->entries, sizeof(log->entries), log_key, sizeof(log_key), log->entries_hmac);
			  passman_chacha20((u8 *)log->entries, sizeof(log->entries), log_key, log->nonce, (u8 *)log->entries);
		  
			  if (!generate_log_hmac(log, log_key, log->hmac))
			    passman_error("list", 1, "cannot compute log HMAC\n");
			  
			  passman_memzero(log_key, sizeof(log_key));
			  
			  if (!write_log(log_file_path, log))
			    {
			      passman_memzero((u8 *)db, sizeof(passman_db_t));
			      passman_memzero((u8 *)log, sizeof(passman_log_t));
			      free(db);
			      free(log);
		      
			      passman_error("list", 1, "cannot write into log\n");
			    }
			  else
			    {
			      passman_memzero((u8 *)db, sizeof(passman_db_t));
			      passman_memzero((u8 *)log, sizeof(passman_log_t));
			      free(db);
			      free(log);

			      return 0;
			    }
			}
		      else
			{
			  passman_memzero((u8 *)db, sizeof(passman_db_t));
			  passman_memzero((u8 *)log, sizeof(passman_log_t));
			  free(db);
			  free(log);
		  
			  passman_error("list", 1, "cannot add entry or write to log\n");
			}
		    }
		  else
		    {
		      passman_memzero((u8 *)db, sizeof(passman_db_t));
		      passman_memzero((u8 *)log, sizeof(passman_log_t));
		      free(db);
		      free(log);
	      
		      passman_error("list", 1, "invalid database or log signature\n");
		    }
		}
	      else
		{
		  passman_memzero((u8 *)db, sizeof(passman_db_t));
		  passman_memzero((u8 *)log, sizeof(passman_log_t));
		  free(db);
		  free(log);
	  
		  passman_error("list", 1, "invalid master password\n");
		}
	    }
	}
    }
  
  if (mode == PASSMAN_MODE_LOG)
    {
      u64 password_len = 0;
      u8 log_key[CHACHA20_KEY_LEN];
      u8 log_hmac[SHA3_256_HASH_LEN];
      ascii password[PASSMAN_MAX_STR];

      passman_log_t *log = read_log(log_file_path);
      
      if (!log)
	passman_error("log", 1, "cannot read log\n");
      else
	{
	  printf("master password: ");
	  get_secret(password, &password_len, PASSMAN_NOECHO);
	  
	  generate_log_key((u8 *)password, password_len, log, log_key);
	  
	  passman_memzero((u8 *)password, sizeof(password)); password_len = 0;
	  
	  if (!generate_log_hmac(log, log_key, log_hmac))
	    passman_error("log", 1, "cannot compute log HMAC\n");
	  
	  if (!passman_memcmp(log_hmac, log->hmac, SHA3_256_HASH_LEN))
	    {
	      passman_chacha20((u8 *)log->entries, sizeof(log->entries), log_key, log->nonce, (u8 *)log->entries);
	      passman_sha3_256_hmac((u8 *)log->entries, sizeof(log->entries), log_key, sizeof(log_key), log_hmac);

	      passman_memzero(log_key, sizeof(log_key));
	      
	      if (!passman_memcmp(log_hmac, log->entries_hmac, SHA3_256_HASH_LEN))
		{
		  if (!check_log(log))
		    {
		      passman_memzero((u8 *)log, sizeof(passman_log_t));		  
		      free(log);
		      
		      passman_error("log", 1, "corrupt log blockchain\n");
		    }
		  else
		    {
		      print_log(log);
		      
		      passman_memzero((u8 *)log, sizeof(passman_log_t));		  
		      free(log);
		      
		      return 0;
		    }
		}
	      else
		{
		  passman_memzero((u8 *)log, sizeof(passman_log_t));
		  free(log);
		  
		  passman_error("log", 1, "invalid log signature\n");
		}
	    }
	  else
	    {
	      passman_memzero((u8 *)log, sizeof(passman_log_t));
	      free(log);
	      
	      passman_error("list", 1, "invalid master password\n");
	    }
	}
    }
}
